/* $Id: et4000_text16.inc,v 1.6 1998/11/24 18:28:09 marcus Exp $
******************************************************************************

   ET4000 text16 operators

   Copyright (C) 1997 Jason McMullan	[jmcc@ggi-project.org]
   Copyright (C) 1997 Andreas Vogler	[duff@allgaeu.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************

	This code was derived from the following source of information:
	[1]     IBM vga driver by Jason McMullan
	[2]     vgadocs4 by Finn Thoegersen

******************************************************************************
*/

#ifndef _ET4000_TEXT16_INC_
#define _ET4000_TEXT16_INC_

#ifndef MAINTAINER
# define MAINTAINER	Gernot_Ziegler
#endif

#undef  DEBUG_P		/* Debug Font-pointer (_very_ verbose!) */

#include "IBM/vga_text16.h"
#include "pointer_mask.h"

#define CURSOR_MASK	0xE8		/* Attribute mask (bg,fg) */
#define POINTER_MASK	0xFF		/* Attribute mask (bg,fg) */

#define PCELLSIDEX	3		/* Horiz.  length of pointer cell box */
#define PCELLSIDEY	3		/* Vertic. length of pointer cell box */

#define PCELLTOTAL	(PCELLSIDEX*PCELLSIDEY)		
#define PCELLOF(fp,xsize,i) ((fp)+(((i)/PCELLSIDEX)*(xsize)) + ((i)%PCELLSIDEX))

typedef struct text16_point_s {
	int		visible;	/* Should I visible the old data? */
	int		offset;		/* Offset of old point (charcells) */
	int		offset2;	/* Extra offset, due to split line */
	int		splity;         /* Split line in old point (0..PCELLSIDEY) */
	kgi_uint		maxx;		/* Max visible x of old point (charcells) */
	kgi_uint		maxy;		/* Max visible y of old point (charcells) */
	int		bitoffset;	/* Offset of old point (dots) */
	charcell	data[PCELLTOTAL];	/* Data at old point */
} text16_point;
	
struct text16_data_s {
	kgi_uint		magic;
	int		font_changed;
	unsigned long	origin;
	int		splitline;
	text16_point  	pointer;
} text16_data[MAX_NR_DISPLAYS];

#define ORIGIN 		(data->origin)
#define SPLITLINE	(data->splitline)
#define FONTCHANGED	(data->font_changed)

#define PVISIBLE	(data->pointer.visible)
#define POFFSET1	(data->pointer.offset)
#define POFFSET2	(data->pointer.offset2)
#define PSPLITY		(data->pointer.splity)
#define PMAXX		(data->pointer.maxx)
#define PMAXY		(data->pointer.maxy)
#define PBITOFFSET	(data->pointer.bitoffset)
#define PDATA		(data->pointer.data)

static void text16_data_init(void)
{
	memset(text16_data,0,sizeof(struct text16_data *)*MAX_NR_DISPLAYS);
}
	
static struct text16_data_s *text16_data_realloc(struct text16_data_s *old)
{
	int i;

	if (old==NULL || old->magic!=TEXT16_MAGIC) {
		for (i=0;i<MAX_NR_DISPLAYS;i++) 
			if (text16_data[i].magic!=TEXT16_MAGIC) {
			  memset(&(text16_data[i]),0,
				 sizeof(struct text16_data_s));
			  text16_data[i].magic=TEXT16_MAGIC;
			  return &(text16_data[i]);
			}
		return NULL;
	}
	memset(old,0,sizeof(struct text16_data_s));
	old->magic=TEXT16_MAGIC;
	return old;
}

/*
 Text-font based mouse pointer code...
*/

/* Ascii codes of font locs. to
 * abuse for our purposes..
 */
static kgiu8 pointer_char[PCELLTOTAL]={
	0xC5,0xC6,0xC7,
	0xD6,0xD7,0xD8,
	0xCF,0xDD,0xDE
};

#ifdef DEBUG_P
static void print_mask(const char *msg,kgi_uint *mask,kgi_uint texty)
{
  int i,j;
  char buff[33];

  printk(KERN_DEBUG"%s\n",msg);

  buff[32]=0;
  for (i=0;i<texty*2;i++) {
    for (j=0;j<32;j++) 
      buff[31-j]=(mask[i]&(1<<j)) ? 'O' : ' ';    
    printk(KERN_DEBUG"  [%.2d]%s\n",i,buff);
  }
}
#endif /* DEBUG_P */

typedef struct text16_font_reg_s {
	kgiu8	SR2,SR4;
	kgiu8	GR4,GR5,GR6;
} text16_font_reg;

#define FONT_GET 0
#define FONT_SET 1

static inline void text16_font_enter(text16_font_reg *regs) 
{
    IS_CRITICAL;

#ifdef SLOW_POINTER
    WAIT_VRETRACE();
#endif
    ENTER_CRITICAL;
    regs->SR2 = inbSEQ(0x02);
    regs->SR4 = inbSEQ(0x04);		/* save register state        */
    regs->GR4 = inbGRC(0x04);
    regs->GR5 = inbGRC(0x05);
    regs->GR6 = inbGRC(0x06);

    outbSEQ(SR02_PLANE2, 0x02);	/* font data goes to plane 2  */
    outbSEQ(SR04_256K_ACCESS | SR04_NO_ODDEVEN, 0x04);

    outbGRC(2, 0x04);		/* read plane 2               */
    outbGRC(GR5_WRITEMODE0, 0x05);
    outbGRC(GR6_MAP_A0_64 | GR6_GRAPHMODE, 0x06);
    LEAVE_CRITICAL;
}

static void inline text16_font_leave(text16_font_reg *regs)
{
    IS_CRITICAL;

    ENTER_CRITICAL;
    outbSEQ(regs->SR2, 0x02);
    outbSEQ(regs->SR4, 0x04);		/* Restore registers  */
    outbGRC(regs->GR4, 0x04);
    outbGRC(regs->GR5, 0x05);
    outbGRC(regs->GR6, 0x06);
    LEAVE_CRITICAL;
}

static inline void direct_text16_font_read(int chr,kgiu8 *charcell)
{
    memcpy(charcell,(kgiu8 *)(phys_to_virt(0xA0000))+(chr<<5),32);
}

static inline void direct_text16_font_write(int chr,kgiu8 *charcell)
{
    memcpy((kgiu8 *)(phys_to_virt(0xA0000))+(chr<<5),charcell,32);
}

static void place_pointer(charcell *fb,kgi_uint xsize,kgi_uint dx,kgi_uint dy,
			  kgi_uint texty,int isdelt,struct text16_data_s *data)
{
    int i,j;
    kgi_uint mask[48];
    kgiu8 font[PCELLTOTAL*32],*cp;
    text16_font_reg regs;
    charcell *fb1=fb+POFFSET1, *fb2=fb+POFFSET2;
    charcell *cell=PDATA;
    kgi_uint maxx=PMAXX, maxy=PMAXY;
    int splity=PSPLITY;
    
    if (dy>16) {
	printk(KERN_DEBUG"PtrHack: Can't deal with characters >16 bits high\n");
	return;
    }
   
#define LOOPOVERxyi { register int x,y; for (y=0;y<maxy;y++) { \
			i=y*PCELLSIDEX; \
			fb=(y<splity) ? fb1 : fb2; \
			for (x=0;x<maxx;x++,i++) {
#define ENDLOOPxyi  }}}

    /* Has anything changed since the last time we 
     * mucked with the fonts? 
     */ 
    if (!isdelt) {
	LOOPOVERxyi
	    isdelt |= (cell[i] != *PCELLOF(fb,xsize,i));
	ENDLOOPxyi
		    
    /* Nope - same characters as we used last time..
     * Why redo our effort?
     */
		    
	if (!isdelt) {
	    LOOPOVERxyi
		CHARSET(PCELLOF(fb,xsize,i),pointer_char[i]);
	    ENDLOOPxyi
	    return;
	}
    } 

    /* Get the characters cells and 
     * their associated fonts..
     */
    LOOPOVERxyi
	cell[i] = *PCELLOF(fb,xsize,i);	
    ENDLOOPxyi

    text16_font_enter(&regs);

    LOOPOVERxyi
	direct_text16_font_read(CHAR(cell[i]),font+(i<<5));
    ENDLOOPxyi

    /* Copy the old chars, as  a bitmap, 
     * into mask[]
     */
    memset(mask,0,sizeof(mask)); 
    LOOPOVERxyi
	cp=font+(i<<5);
	for (j=0;j<texty;j++) 
		mask[j+y*texty] |= (*(cp++)) << ((3-x)*8);
    ENDLOOPxyi

#ifdef DEBUG_P
    print_mask("Original",mask,texty);
#endif

    /* Mask with the And and Or masks... 
     */
    for (i=0;i<POINTER_MASK_LEN;i++) {
	mask[dy+i] |= (pointer_or_mask[i] >> dx);
	mask[dy+i] ^= (pointer_xor_mask[i] >> dx);
    }
         
#ifdef DEBUG_P
    print_mask("W/Pointer",mask,texty);
#endif

    /* And write the new data... */
    LOOPOVERxyi
	cp=font+(i<<5);
	for (j=0;j<texty;j++) 
		*(cp++) = (mask[j+y*texty] >> ((3-x)*8)) & 0xFF;
    ENDLOOPxyi


    LOOPOVERxyi
	direct_text16_font_write(pointer_char[i],font+(i<<5));
    ENDLOOPxyi

    text16_font_leave(&regs);

    LOOPOVERxyi
	CHARSET(PCELLOF(fb,xsize,i),pointer_char[i]);
    ENDLOOPxyi

}

/*
 Do not #undef LOOPOVERxyi and ENDLOOPxyi yet!
 We need them for erase_pointer()
*/

static void erase_pointer(charcell *fb,kgi_uint xsize,struct text16_data_s *data)
{
    int i;
    charcell *fb1=fb+POFFSET1, *fb2=fb+POFFSET2;
    kgi_uint maxx = PMAXX, maxy = PMAXY;
    int splity = PSPLITY;

    LOOPOVERxyi
	*PCELLOF(fb,xsize,i)=PDATA[i];
    ENDLOOPxyi
}

#undef LOOPOVERxyi
#undef ENDLOOPxyi

/*
 Font setting
*/

static void text16_do_set_font(int no,int from,int to,struct kgi_font *font)
{
    text16_font_reg fontregs;
    int i;
    char *fbuf, *data;
    register int cnt = 32 - font->dy;

    if (font == NULL)
	return;

    fbuf = (char *) phys_to_virt(0x000A0000);
    fbuf += (no * 0x2000) + (from << 5);
    data = font->data+font->dy*from;
    if (!data)
      return;

    text16_font_enter(&fontregs);
    /* NOTE: no printk()'s are allowed in here unless klogd is active !!!  */

    for (i = from; i <= to; i++) {
    	memmove(fbuf, data, font->dy);
	memset(fbuf + font->dy, 0, cnt);
	fbuf += 32;
	data += font->dy;
    };

    text16_font_leave(&fontregs);

    outbCRT((inbCRT(0x0A) & ~CR0A_CURSOR_START_MASK)	/* setup cursor    */
     	     | ((font->base+1) & CR0A_CURSOR_START_MASK), 0x0A);
    outbCRT((inbCRT(0x0B) & ~CR0B_CURSOR_END_MASK)
   	     | ((font->base+2) & CR0B_CURSOR_END_MASK), 0x0B);
}

static void text16_set_font(struct kgi_display *dpy,
                            struct kgi_font *font,kgi_uint from,kgi_uint to)
{
  struct text16_data_s *data = (struct text16_data_s *)(dpy->driver_data);
  text16_do_set_font(0,from,to,font);

  FONTCHANGED=1;
}

static void text16_set_origin(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	struct text16_data_s *data = (struct text16_data_s *)(dpy->driver_data);
	IS_CRITICAL;
	static kgi_uint last_xdot = -1;
	static kgi_uint last_ydot = -1;
	register kgi_uint xpix = x / dpy->mode.request.dpp.x;
	register kgi_uint xdot = x % dpy->mode.request.dpp.x;
	register kgi_uint ypix = y / dpy->mode.request.dpp.y;
	register kgi_uint ydot = y % dpy->mode.request.dpp.y;

	register kgi_uint offset = xpix + ypix*dpy->fb.size.x;

	ENTER_CRITICAL;

		if (ORIGIN != offset) {

			outbCRT(offset >> 8, CRT_STARTADDR_H);
			outbCRT(offset,      CRT_STARTADDR_L);
			if(chip==ET4000) {
				offset &= 0x3FFFF;
				outb((inbCRT(CRT_EXT_ADDR) &
				      ~CR33_DISPLAYSTART_MASK)
				     | (offset >> 16), CRT_EXT_ADDR);
			} else {
				outb((inbCRT(CRT_EXT_ADDR) &
				      ~CR33_DS_MASK_W32X)
				     | (offset >> 16), CRT_EXT_ADDR);
			}
			ORIGIN = offset;
		}

		if (last_ydot != ydot) {

			outbCRT(ydot & CR08_ROWSCAN_MASK, CRT_PRESETROW);
			last_ydot = ydot;
		}

		if (last_xdot != xdot) {

			register kgi_uint hpan;

			if (dpy->mode.request.dpp.x == 9) {

				hpan = (xdot == 0)
					    ? 0x08 : (xdot-1) & AR13_DATA_MASK;
			} else {

				hpan = xdot & AR13_DATA_MASK;
			}
			WAIT_HVRETRACE();
			outbATC(hpan, ATC_HPAN);
			outb( ATCI_ENABLEDISPLAY, ATCI);
			last_xdot = xdot;
		}

	LEAVE_CRITICAL;
}

static void text16_set_split(struct kgi_display *dpy, int line)
{
	struct text16_data_s *data = (struct text16_data_s *)(dpy->driver_data);
	register kgiu8 foo;

	outbCRT(line & 0xFF, CRT_LINECOMPARE);

	foo = (line & 0x0100)
		? inbCRT(0x07) | CR07_LINECOMPARE_B8
		: inbCRT(0x07) & ~CR07_LINECOMPARE_B8;
	outbCRT(foo, 0x07);

	foo = (line & 0x0200)
		? inbCRT(0x09) | CR09_LINECOMPARE_B9
		: inbCRT(0x09) & ~CR09_LINECOMPARE_B9;
	outbCRT(foo, 0x09);

	foo = (line & 0x0400)
		? inbCRT(0x35) | CR35_LINECOMPARE_B10
		: inbCRT(0x35) & ~CR35_LINECOMPARE_B10;
	outbCRT(foo, 0x35);

	SPLITLINE=line;
}

#define	text16_set_frame	NULL	/* !!! implement! */

#define DPPX           (dpy->mode.request.dpp.x)
#define DPPY           (dpy->mode.request.dpp.y)
#define VIRTUALX        (dpy->fb.size.x)
#define VIRTUALY        (dpy->fb.size.y)
#define VISIBLEX        (dpy->mode.request.visible.x)
#define VISIBLEY        (dpy->mode.request.visible.y)

void text16_show_pointer(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	struct text16_data_s *data = (struct text16_data_s *)(dpy->driver_data);
        kgiu32 newoff1 = ORIGIN + (x / DPPX)  +  (y / DPPY) * VIRTUALX;
	kgiu32 newoff2 = 0;
        register charcell *p = (void *) dpy->fb.pbuf;
	kgi_uint bitoff = x+y*VISIBLEX;
	kgi_uint dx,dy,clipx,clipy,maxx,maxy;
	int splity = PCELLSIDEY-1;
	
	if (SPLITLINE && SPLITLINE < VIRTUALY*DPPY &&
	    y+POINTER_MASK_LEN > SPLITLINE) {
		newoff2=(x / DPPX) + ((y-SPLITLINE) / DPPY) * VIRTUALX;
		if (y >= SPLITLINE) {
			newoff1 = newoff2;
			newoff2 = 0;
		}
		else 
			splity = SPLITLINE/DPPY - y/DPPY + 1;
	}

#ifdef DEBUG_P
	printk(KERN_DEBUG"PShow(%c):POFFSET1=0x%.8x POFFSET2=0x%.8x PBOFFSET=0x%.8x x=%d,y=%d\n",
		PVISIBLE ? 'V' : 'i',POFFSET1,POFFSET2,PBITOFFSET,x,y);
#endif

 	ASSERT((x < VISIBLEX *DPPX) && (y < VISIBLEY * DPPY));

	if (PVISIBLE) 
		erase_pointer(p,VIRTUALX,data);

	dx=x % DPPX;
	dy=y % DPPY;
	clipx = VIRTUALX-(x/DPPX);
	clipy = VIRTUALY-(y/DPPY);

	maxx=((POINTER_MASK_WIDTH-1+dx)/8)+1;
	maxx=(maxx > clipx) ? clipx : maxx;
	maxy=((POINTER_MASK_LEN-1+dy)/DPPY)+1;
	maxy=(maxy > clipy) ? clipy : maxy;

	if (splity<0)
		splity = 0;
	else if (splity>maxy)
		splity = maxy;

        POFFSET1=newoff1;
	POFFSET2=newoff2;
	PSPLITY=splity;
	PMAXX=maxx;
	PMAXY=maxy;
	
	place_pointer(p,VIRTUALX,dx,dy,DPPY,bitoff!=PBITOFFSET || FONTCHANGED,
		      data);

	PBITOFFSET=bitoff;
	PVISIBLE=1;
	FONTCHANGED=0;

}

static void text16_hide_pointer(struct kgi_display *dpy)
{
	struct text16_data_s *data = (struct text16_data_s *)(dpy->driver_data);
        register charcell *p = (void *) dpy->fb.pbuf;

#ifdef DEBUG_P
	printk(KERN_DEBUG"PHide(%c): POFFSET1=0x%.8x POFFSET2=0x%.8x PBOFFSET=0x%.8x\n",
		PVISIBLE ? 'V' : 'i',POFFSET1,POFFSET2,PBITOFFSET);
#endif

	if (PVISIBLE) {
		erase_pointer(p,VIRTUALX,data);
		PVISIBLE=0;
	}
}

#define text16_undo_pointer     text16_hide_pointer

static void text16_show_cursor(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
       	struct text16_data_s *data = (struct text16_data_s *)(dpy->driver_data);
        register kgiu32 new = ORIGIN + (x / DPPX)  +  (y / DPPY) * VIRTUALX;
 	IS_CRITICAL;

        ASSERT((x < VISIBLEX * DPPX) && (y < VISIBLEY * DPPY));

	ENTER_CRITICAL;
	outbCRT(new >> 8, CRT_CURSORADDR_H);
	outbCRT(new&0xFF, CRT_CURSORADDR_L);
	if(chip==ET4000) {		
		outbCRT((inbCRT(CRT_EXT_ADDR) & ~CR33_CURSORSTART_MASK)
			| ((new & 0x30000) >> 14), CRT_EXT_ADDR);
	} else {
		outbCRT((inbCRT(CRT_EXT_ADDR) & ~CR33_CS_MASK_W32X)
			| ((new & 0xF0000) >> 12), CRT_EXT_ADDR);
	}
	LEAVE_CRITICAL;
}

static void text16_hide_cursor(struct kgi_display *dpy)
{
	IS_CRITICAL;

	ENTER_CRITICAL;
	outbCRT(0xFF,CRT_CURSORADDR_H);
	outbCRT(0xFF,CRT_CURSORADDR_L);
	if(chip==ET4000) {		
		outbCRT((inbCRT(CRT_EXT_ADDR) | CR33_CURSORSTART_MASK),
			CRT_EXT_ADDR);
	} else {
		outbCRT((inbCRT(CRT_EXT_ADDR) | CR33_CS_MASK_W32X),
			CRT_EXT_ADDR);
	}
	LEAVE_CRITICAL;
}

/*
 This is badly wrong. Hardware cursor doesn't need and undo function!
 We should put NULL here and update kgi.c to actually check if dpy->undo_xxxx
 are not NULL before calling them.
*/

/* I guess it has been changed in kgi.c by now */

#define text16_undo_cursor     NULL 

#undef SPLITLINE
#undef ORIGIN

#undef PVISIBLE
#undef POFFSET1
#undef POFFSET2
#undef PSPLITY
#undef PMAXX
#undef PMAXY
#undef PDATA

#endif /* _ET4000_TEXT16_INC */
